
(*
 dual vector
 vecteur dual
 双対ベクトル
  
 *)

Section F_dual.

  Variable F : Type.
  Variable add mul : F->F->F.
  Axiom assoc_add : forall (a b c : F), add a (add b c) = add (add a b) c.
  Axiom assoc_mul : forall (a b c : F), mul a (mul b c) = mul (mul a b) c.
  Axiom comm_add : forall (a b : F), add a b = add b a.
  Axiom comm_mul : forall (a b : F), mul a b = mul b a.
  Axiom distribute_left : forall (a b c : F), mul a (add b c) = add (mul a b) (mul a c).
  Axiom distribute_right : forall (a b c : F), mul (add a b) c = add (mul a c) (mul b c). 
  
  (* vector の和 *)
  Definition add_v (u v : F*F) : F*F :=
    let (u1, u2) := u in
    let (v1, v2) := v in
    (add u1 v1, add u2 v2).

  (* Dual vector の和 *)
  Definition add_f (f g : F*F -> F) (u : F*F) : F := add (f u) (g u).

  Lemma add_ext1 : forall (a1 a2 a3 a4 : F), add (add a1 a2) (add a3 a4) = add (add (add a1 a2) a3) a4.
  Proof.
    intros.
    rewrite assoc_add.
    reflexivity.
  Qed.

  Lemma add_ext2 : forall (a1 a2 a3 a4 : F), add (add a1 a2) a3 = add (add a1 a3) a2.
  Proof.
    intros.
    repeat rewrite <- assoc_add.
    assert (add a2 a3 = add a3 a2).
    - rewrite comm_add.
      reflexivity.
    - rewrite H.
      reflexivity.
  Qed.

  Lemma add_ext3 : forall (a1 a2 a3 a4 : F), add (add a1 a2) (add a3 a4) = add (add a1 a3) (add a2 a4).
  Proof.
    intros.
    repeat rewrite assoc_add.
    assert (add (add a1 a2) a3 = add (add a1 a3) a2).
    - repeat rewrite <- assoc_add.
      + assert (add a2 a3 = add a3 a2).
        rewrite comm_add.
        reflexivity.
      rewrite H.
      reflexivity.
    - rewrite H.
      reflexivity.
  Qed.

  Lemma add_dual_linear : forall (u v : F*F) (f g : F*F->F),
                                     f (add_v u v) = add (f u) (f v) ->
                                     g (add_v u v) = add (g u) (g v) ->
                                     (add_f f g) (add_v u v) = add ((add_f f g) u) ((add_f f g) v).
Proof.
  intros u v f g Hf Hg.
  unfold add_f.
  rewrite Hf.
  rewrite Hg.
  rewrite add_ext3.
  reflexivity.
Qed.

(* vector のスカラー倍 *)
Definition mul_v (c : F) (u : F*F) : F*F :=
  let (u1, u2) := u in
  (mul c u1, mul c u2).

(* Dual vector のスカラー倍 *)
Definition mul_f (c : F) (f : F*F -> F) (u : F*F) : F := mul c (f u).


Lemma mul_dual_linear : forall (c0 c1 : F) (u : F*F) (f : F*F->F),
                          f (mul_v c1 u) = mul c1 (f u) ->
                          (mul_f c0 f) (mul_v c1 u) = mul (mul c0 c1) (f u). 
Proof.
  intros c0 c1 u f Hv.
  unfold mul_f.
  rewrite Hv.
  rewrite assoc_mul.
  reflexivity.
Qed.

Definition bilinear_form (u v : F*F) : F :=
  let (u1, u2) := u in
  let (v1, v2) := v in
  add (mul u1 v1) (mul u2 v2).

Definition dual (u : F*F) : F*F -> F := bilinear_form u.

Lemma dual_bilinear : forall (u1 u2 v1 v2 : F), (dual (u1,u2)) (v1,v2) = add (mul u1 v1) (mul u2 v2).
Proof.
  intros.
  unfold dual.
  unfold bilinear_form.
  reflexivity.
Qed.



Lemma dual_map_add : forall (u1 u2 v1 v2 v1' v2': F), (dual (u1,u2)) (add_v (v1,v2) (v1',v2')) = add ((dual (u1,u2)) (v1,v2)) ((dual (u1,u2)) (v1',v2')).
Proof.
  intros.
  unfold dual.
  unfold add_v.
  unfold bilinear_form.
  repeat rewrite distribute_left.
  rewrite add_ext3.
  reflexivity.
Qed.

Lemma dual_map_mul : forall (c u1 u2 v1 v2 : F), (dual (u1,u2)) (mul_v c (v1,v2)) = mul c ((dual (u1,u2)) (v1,v2)).
Proof.
  intros.
  unfold dual.
  unfold mul_v.
  unfold bilinear_form.
  rewrite distribute_left.
  repeat rewrite assoc_mul.
  assert (mul u1 c = mul c u1).
  - rewrite comm_mul.
    reflexivity.
  - assert (mul u2 c = mul c u2).
    + rewrite comm_mul.
      reflexivity.
    + rewrite H.
      rewrite H0.
      reflexivity.
Qed.      

Theorem dual_vector_add : forall (u1 u2 u1' u2' v1 v2 : F), dual (add_v (u1,u2) (u1', u2')) (v1, v2) = add (dual (u1,u2) (v1,v2)) (dual (u1',u2') (v1,v2)). 
Proof.
  intros.
  unfold dual.
  unfold add_v.
  unfold bilinear_form.
  repeat rewrite distribute_right.
  rewrite add_ext3.
  reflexivity.
Qed.

Theorem dual_vector_mul : forall (c u1 u2 v1 v2 : F), dual (mul_v c (u1,u2)) (v1,v2) = mul c (dual (u1,u2) (v1,v2)).
Proof.
  intros.
  unfold dual.
  unfold mul_v.
  unfold bilinear_form.
  rewrite distribute_left.
  repeat rewrite assoc_mul.
  reflexivity.
Qed.
End F_dual.

Extraction "dual_complete.ml" add_v add_f mul_v mul_f bilinear_form dual.